﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Threading;

namespace EonaCat.Sockets
{
    public class SocketServer : IDisposable
    {
        /// <summary>
        /// Unique server id.
        /// </summary>
        public string Guid { get; set; }

        /// <summary>
        /// Listening ip address.
        /// </summary>
        public IPAddress IPAddress { get; set; }

        /// <summary>
        /// Listening port.
        /// </summary>
        public int Port { get; set; }

        /// <summary>
        /// List of connected clients.
        /// </summary>
        public List<ClientConnection> ConnectedClients { get; set; } = new List<ClientConnection>();

        /// <summary>
        /// Interval to check for disconnected clients
        /// </summary>
        public int DisconnectionCheckInterval { get; set; } = 100;

        /// <summary>
        /// Determines if we need to reuse the socket address
        /// (default: false)
        /// </summary>
        public bool ReuseAddress { get; set; }

        /// <summary>
        /// Determines if the socket is in blocking mode
        /// (default: true)
        /// </summary>
        public bool IsSocketInBlockingMode { get; set; } = true;


        /// <summary>
        /// Socket backlog. The maximum length of the pending connections queue.
        /// </summary>
        public int Backlog { get; set; } = 0;
        /// <summary>
        /// Server type.
        /// </summary>
        public NetworkType ServerType { get; set; }

        /// <summary>
        /// Only for UDP.
        /// Required to use SocketClient class for this.
        /// </summary>
        public bool UDPClientManage { get; set; } = true;

        /// <summary>
        /// Only for UDP.
        /// Accept 1 byte data for connected state check.
        /// </summary>
        private int _UDPDataInterval { get => (int)(UDPDataInterval * 1.5); }
        public int UDPDataInterval { get; set; } = 5000;

        /// <summary>
        /// Called when new client connected
        /// </summary>
        public event EventHandler<SocketServerClientEventArgs> OnConnected;
        /// <summary>
        /// Called when client disconnected
        /// </summary>
        public event EventHandler<SocketServerClientEventArgs> OnDisconnected;
        /// <summary>
        /// Called when received data from client
        /// </summary>
        public event EventHandler<SocketServerDataEventArgs> OnReceived;
        /// <summary>
        /// Called when server catch error
        /// </summary>
        public event EventHandler<ErrorEventArgs> OnError;
        /// <summary>
        /// Called when executing Start() function
        /// </summary>
        public event EventHandler<string> OnStart;
        /// <summary>
        /// Called when executing Stop() function
        /// </summary>
        public event EventHandler OnStop;

        private Socket Listener { get; set; }
        private System.Timers.Timer DisconnectTimer { get; set; }
        private Dictionary<EndPoint, double> LastDataReceivedTime { get; set; }
        private double TimeNow { get => (DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds; }

        public SocketServer(NetworkType type, string address, int port)
        {
            if (IPAddress.TryParse(address, out var ipAddress))
            {
                new SocketServer(type, ipAddress, port);
            }
            throw new Exception("Invalid IP address specified");
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="type">Server type (TCP, UDP)</param>
        /// <param name="address">Server listening ip</param>
        /// <param name="port">Server listening port</param>
        public SocketServer(NetworkType type, IPAddress address, int port)
        {
            Guid = System.Guid.NewGuid().ToString();
            ServerType = type;
            IPAddress = address;
            Port = port;
            DisconnectTimer = new System.Timers.Timer(DisconnectionCheckInterval);
        }

        public void Dispose()
        {
            Stop();
        }

        /// <summary>
        /// Start server
        /// </summary>
        public void Start()
        {
            try
            {
                switch (ServerType)
                {
                    case NetworkType.Tcp:
                        Listener = new Socket(IPAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
                        ConfigureSocketOptions();

                        // Check if we need to override the socket with a custom socket
                        if (CustomTcpSocket != null)
                        {
                            Listener = CustomTcpSocket;
                        }

                        DisconnectTimer.Elapsed += (s, e) =>
                        {
                            lock (ConnectedClients)
                            {
                                ConnectedClients.RemoveAll(x =>
                                {
                                    if (x.IsConnected)
                                    {
                                        return false;
                                    }

                                    OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = x });
                                    return true;
                                });
                            }
                        };
                        break;
                    case NetworkType.Udp:
                        {
                            if (UDPClientManage)
                            {
                                LastDataReceivedTime = new Dictionary<EndPoint, double>();
                                DisconnectTimer.Elapsed += (s, e) =>
                                {
                                    var now = TimeNow;
                                    var times = LastDataReceivedTime;
                                    var removed = new List<EndPoint>();
                                    foreach (var kp in times)
                                    {
                                        if (now - kp.Value > _UDPDataInterval / 1000)
                                        {
                                            lock (ConnectedClients)
                                            {
                                                var client = ConnectedClients.Where(x => x.Guid == kp.Key.ToString());
                                                if (!client.Any())
                                                {
                                                    continue;
                                                }
                                                OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client.First() });
                                                ConnectedClients.Remove(client.First());
                                                removed.Add(kp.Key);
                                            }
                                        }
                                    }
                                    lock (LastDataReceivedTime)
                                    {
                                        foreach (var r in removed)
                                        {
                                            LastDataReceivedTime.Remove(r);
                                        }
                                    }
                                };
                            }
                            Listener = new Socket(IPAddress.AddressFamily, SocketType.Dgram, ProtocolType.Udp);
                            ConfigureSocketOptions();

                            // Check if we need to override the socket with a custom socket
                            if (CustomUdpSocket != null)
                            {
                                Listener = CustomUdpSocket;
                            }

                            break;
                        }
                }
                Listener.ReceiveBufferSize = ClientConnection.BufferSize;
                Listener.SendBufferSize = ClientConnection.BufferSize;
                Listener.Bind(new IPEndPoint(IPAddress, Port));
                if (ServerType == NetworkType.Tcp)
                {
                    Listener.Listen(Backlog);
                }
                DisconnectTimer.Start();
                ListenerLoop();
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        /// <summary>
        /// Setup this udp socket to override the default one
        /// (default: null => using default)
        /// </summary>
        public Socket CustomUdpSocket { get; set; }

        /// <summary>
        /// Setup this tcp socket to override the default one
        /// (default: null => using default)
        /// </summary>
        public Socket CustomTcpSocket { get; set; }

        static void SetConnectionReset(Socket socket)
        {
            try
            {
                if (socket.ProtocolType == ProtocolType.Udp && RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                {
                    // Disable ICMP packet shutdown (forcibly closed)
                    const uint IOC_IN = 0x80000000;
                    const uint IOC_VENDOR = 0x18000000;
                    uint SIO_UDP_CONNRESET = IOC_IN | IOC_VENDOR | 12;
                    socket.IOControl((int)SIO_UDP_CONNRESET, new[] { Convert.ToByte(false) }, null);
                };
            }
            catch { }
        }

        private void ConfigureSocketOptions()
        {
            // Disable ICMP packet shutdown
            SetConnectionReset(Listener);

            if (ReuseAddress)
            {
                Listener.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            }

            Listener.Blocking = IsSocketInBlockingMode;
        }

        /// <summary>
        /// Stop server
        /// </summary>
        public void Stop()
        {
            try
            {
                DisconnectTimer.Stop();
                foreach (var client in ConnectedClients)
                {
                    OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                    client.Socket.Close();
                }
                ConnectedClients.Clear();
                OnStop?.Invoke(this, EventArgs.Empty);
                Listener.Close();
            }
            catch (SocketException socketException)
            {
                OnError?.Invoke(this, new ErrorEventArgs(socketException));
            }
        }

        private void ListenerLoop()
        {
            try
            {
                OnStart?.Invoke(this, CustomTcpSocket != null || CustomUdpSocket != null ? "Custom socket was used" : string.Empty);
                if (ServerType == NetworkType.Tcp)
                {
                    Listener.BeginAccept(AcceptCallback, Listener);
                }

                if (ServerType != NetworkType.Udp)
                {
                    return;
                }
                var loopbackAddress = Listener.AddressFamily == AddressFamily.InterNetworkV6
                    ? IPAddress.IPv6Loopback
                    : IPAddress.Loopback;

                if (UDPClientManage)
                {
                    var buffer = new byte[ClientConnection.BufferSize];
                    EndPoint remoteIp = new IPEndPoint(loopbackAddress, 0);
                    Listener.BeginReceiveFrom(buffer, 0, ClientConnection.BufferSize, 0, ref remoteIp, AcceptCallbackUDP, new object[] { buffer, remoteIp, Listener });
                }
                else
                {
                    var buffer = new byte[ClientConnection.BufferSize];
                    EndPoint remoteEp = new IPEndPoint(loopbackAddress, 0);
                    Listener.BeginReceiveFrom(buffer, 0, ClientConnection.BufferSize, 0, ref remoteEp, ReadCallbackUDP, new object[] { buffer, remoteEp, Listener });
                }
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        private void AcceptCallback(IAsyncResult ar)
        {
            try
            {
                var listener = (Socket)ar.AsyncState;
                var handler = listener.EndAccept(ar);
                var client = new ClientConnection()
                {
                    Guid = System.Guid.NewGuid().ToString(),
                    Socket = handler
                };
                lock (ConnectedClients)
                {
                    ConnectedClients.Add(client);
                }
                OnConnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                client.Socket.BeginReceive(client.Buffer, 0, ClientConnection.BufferSize, 0, ReadCallback, client);
                listener.BeginAccept(AcceptCallback, Listener);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void ReadCallback(IAsyncResult ar)
        {
            try
            {
                var client = (ClientConnection)ar.AsyncState;
                var bytesRead = client.Socket.EndReceive(ar);
                if (bytesRead > 0)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(client.Buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs() { Client = client, Data = data });
                }
                Thread.Sleep(DisconnectionCheckInterval);
                client.Socket.BeginReceive(client.Buffer, 0, ClientConnection.BufferSize, 0, ReadCallback, client);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void AcceptCallbackUDP(IAsyncResult ar)
        {
            try
            {
                var objects = (object[])ar.AsyncState;
                var buffer = objects[0] as byte[];
                var remoteIp = objects[1] as EndPoint;
                var listener = objects[2] as Socket;
                var bytesRead = listener.EndReceiveFrom(ar, ref remoteIp);
                var clients = ConnectedClients.Where(x => x.Guid == remoteIp.ToString()).ToList();
                ClientConnection client;
                if (clients.Count == 0)
                {
                    client = new ClientConnection()
                    {
                        Guid = remoteIp.ToString(),
                        Socket = new Socket(remoteIp.AddressFamily, SocketType.Dgram, ProtocolType.Udp)
                    };
                    client.SendData += SendData;
                    client.Socket.Connect(remoteIp);
                    if (client.IsConnected)
                    {
                        ConnectedClients.Add(client);
                        OnConnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                    }
                }
                else
                {
                    client = clients.First();
                }
                if (bytesRead > 1)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs() { Client = client, Data = data, });
                }
                else
                {
                    if (LastDataReceivedTime.ContainsKey(remoteIp))
                    {
                        LastDataReceivedTime[remoteIp] = TimeNow;
                    }
                    else
                    {
                        LastDataReceivedTime.Add(remoteIp, TimeNow);
                    }
                }
                Listener.BeginReceiveFrom(buffer, 0, ClientConnection.BufferSize, 0, ref remoteIp, AcceptCallbackUDP, new object[] { buffer, remoteIp, Listener });
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void ReadCallbackUDP(IAsyncResult ar)
        {
            try
            {
                var objects = (object[])ar.AsyncState;
                var buffer = objects[0] as byte[];
                var remote_ip = objects[1] as EndPoint;
                var listener = objects[2] as Socket;
                var bytesRead = listener.EndReceiveFrom(ar, ref remote_ip);
                if (bytesRead > 0)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs() { Data = data });
                }
                Listener.BeginReceiveFrom(buffer, 0, ClientConnection.BufferSize, 0, ref remote_ip, ReadCallbackUDP, new object[] { buffer, remote_ip, Listener });
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void DisconnectCallback(IAsyncResult ar)
        {
            try
            {
                var client = (ClientConnection)ar.AsyncState;
                client.Socket.EndDisconnect(ar);
                OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                lock (ConnectedClients)
                {
                    ConnectedClients.Remove(client);
                }
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        public void SendData(byte[] data, EndPoint point)
        {
            try
            {
                Listener.BeginSendTo(data, 0, data.Length, 0, point, EndSendTo, null);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        private void EndSendTo(IAsyncResult ar)
        {
            try
            {
                Listener.EndSendTo(ar);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }
    }
}